import math
import torch
import torch.nn as nn
import torch.nn.functional as F
from torch.distributions.bernoulli import Bernoulli
from torch.distributions.relaxed_bernoulli import RelaxedBernoulli
from torch.distributions.one_hot_categorical import OneHotCategorical
from torch.distributions.relaxed_categorical import RelaxedOneHotCategorical

__all__ = ['SubsetSelect']

class AutoRegress(nn.Module):
    def __init__(self, in_dim, hidden_dim):
        super(AutoRegress, self).__init__()
        self.in_dim = in_dim
        self.hidden_dim = hidden_dim

        self.fc = nn.Sequential(
                nn.Linear(in_features=in_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=1),
                nn.BatchNorm1d(num_features=1),
                nn.Sigmoid(),
                )
        
    def forward(self, hidden, candidate_mask, core_mask):
        B, W, S, H = hidden.shape
        core_mask = core_mask / (core_mask.max() + 1e-6)
        hidden = hidden.view(-1, H)
        p = self.fc(hidden).view(B, W, S, -1)
        
        t = torch.sum(p*candidate_mask*(1.0 - core_mask), dim=2, keepdims=True)
        t = (t>0.5).type(torch.FloatTensor).detach().to(p.device)
        p = p * candidate_mask * (1.0 - t*core_mask)
        p = 1e-10 + (1 - 2e-10)*p
        p = p / (torch.sum(p, dim=2, keepdims=True) + 1e-10)
        p = p.squeeze(3)
        
        if self.training:
            temp = torch.Tensor([0.5]).to(hidden.device)
            p_z = RelaxedOneHotCategorical(temp, probs=p)
            z_i = p_z.rsample()
        else:
            z_i = OneHotCategorical(probs=p).sample()
        return z_i.unsqueeze(3)

class AutoRegressiveSelect(nn.Module):
    def __init__(self, in_features, hidden_dim):
        super(AutoRegressiveSelect, self).__init__()
        self.in_features = in_features
        self.hidden_dim = hidden_dim
        
        self.encoder1 = nn.Sequential(
                nn.Linear(in_features=in_features, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                )
        
        self.encoder2 = nn.Sequential(
                nn.Linear(in_features=in_features, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                )
        
        self.encoder3 = nn.Sequential(
                nn.Linear(in_features=in_features, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                )
        
        self.autoreg = AutoRegress(in_dim=3*hidden_dim, hidden_dim=hidden_dim)

    def forward(self, points, candidate_mask, num_points=2):
        hidden = self.encoder1(points)
        repres = torch.sum(hidden * candidate_mask, dim=2, keepdims=True) / (torch.sum(candidate_mask, dim=2, keepdims=True) + 1e-10)
        repres = repres.repeat(1, 1, points.shape[2], 1)
        
        hidden = torch.cat([self.encoder2(points), repres], dim=3)
        
        hidden_points = self.encoder3(points)
        z = torch.zeros_like(candidate_mask).to(candidate_mask.device)
        z_real = torch.zeros_like(candidate_mask).to(candidate_mask.device)
        
        for i in range(num_points):
            chosen_repres = torch.sum(hidden_points*z, dim=2, keepdims=True) / (torch.sum(z, dim=2, keepdims=True) + 1e-6)
            chosen_repres = chosen_repres.repeat(1, 1, points.shape[2], 1)
            h = torch.cat([hidden, chosen_repres], dim=3)
            z_i = self.autoreg(h, candidate_mask, z)
            z = z + z_i

        return z

class BerDrop(nn.Module):
    def __init__(self, num_gates, in_features=256, reg_scale=1, alpha=1e-3, thres=0.499, construct_real_mask=True):
        super(BerDrop, self).__init__()
        self.num_gates = num_gates
        self.reg_scale = reg_scale
        self.alpha = alpha
        self.thres = thres
        self.construct_real_mask = construct_real_mask
        
        self.q_network = nn.Sequential(
                nn.BatchNorm1d(in_features),
                nn.Linear(in_features=in_features, out_features=in_features//2, bias=False),
                nn.BatchNorm1d(in_features//2),
                nn.ReLU(),
                nn.Linear(in_features=in_features//2, out_features=in_features//2, bias=False),
                nn.BatchNorm1d(in_features//2),
                nn.ReLU(),
                nn.Linear(in_features=in_features//2, out_features=1, bias=False),
                nn.BatchNorm1d(1)
                )
        self.q_network.apply(self.init_weights)
 
    @staticmethod
    def init_weights(m):
        if type(m) == nn.Linear:
            nn.init.xavier_uniform_(m.weight)
    
    def forward(self, x, get_reg=True, return_logits=False):
        batch_size = x.shape[0]
        ways = x.shape[1]
        shots = x.shape[2]
        
        x = x.view(batch_size*ways*shots,-1)
        
        p = torch.sigmoid(self.q_network(x))
        p = p.view(batch_size, ways, shots, self.num_gates)
        p = 1e-10 + (1-2e-10)*p
        
        if return_logits:
            return p
        
        if self.training:
            temp = torch.Tensor([0.5]).to(x.device)
            p_z = RelaxedBernoulli(temp, probs=p)
            z = p_z.rsample()
        else:
            z = Bernoulli(probs=p).sample()
 
        if self.construct_real_mask:
            self.real_mask = Bernoulli(probs=p).sample()
 
        reg = None
        if get_reg:
            reg = p*torch.log(p/self.alpha+1e-10) + (1-p)*torch.log((1-p)/(1-self.alpha)+1e-10)
            reg = reg.sum() * self.reg_scale
        return z, reg

class SubsetSelect(nn.Module):
    def __init__(self, in_features, hidden_dim, alpha=1e-3, num_points=2, construct_real_mask=True):
        super(SubsetSelect, self).__init__()
        self.in_features = in_features
        self.hidden_dim = hidden_dim
        self.num_points = num_points
        self.construct_real_mask = construct_real_mask
        
        self.encoder1 = nn.Sequential(
                nn.Linear(in_features=in_features, out_features=in_features),
                nn.ReLU(),
                nn.Linear(in_features=in_features, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                )
        
        self.encoder2 = nn.Sequential(
                nn.Linear(in_features=in_features, out_features=in_features),
                nn.ReLU(),
                nn.Linear(in_features=in_features, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                nn.ReLU(),
                nn.Linear(in_features=hidden_dim, out_features=hidden_dim),
                )

        self.dbb = BerDrop(num_gates=1, in_features=hidden_dim*2, alpha=alpha, construct_real_mask=construct_real_mask)
        self.autoreg = AutoRegressiveSelect(in_features=in_features, hidden_dim=hidden_dim//2)
        
    def forward(self, D, return_logits=False):
        hidden = self.encoder1(D) 
        
        repres = torch.mean(hidden, dim=2, keepdims=True)
        repres = repres.repeat(1, 1, D.shape[2], 1)
        
        hidden = torch.cat([self.encoder2(D), repres], dim=3)
        
        if return_logits:
            logits = self.dbb(hidden, return_logits=True)
            return logits
        global_mask, kld = self.dbb(hidden)

        if self.training and self.construct_real_mask:
            self.real_mask = self.dbb.real_mask

        local_mask = self.autoreg(points=D.squeeze(1), candidate_mask=global_mask.squeeze(1), num_points=self.num_points).unsqueeze(1)
        return global_mask, local_mask, kld
