import torch
import torch.nn as nn
import torch.nn.functional as F
import torch.optim as optim
from torch.utils.data import DataLoader
from sklearn.metrics import roc_auc_score
from tqdm import tqdm

from celeba import *

import pdb

mode = 'inference'

class Classification_Model(nn.Module):
    def __init__(self):
        super(Classification_Model, self).__init__()
        self.ConvLayer1 = nn.Sequential(
            nn.Conv2d(3, 64, 3), 
            nn.BatchNorm2d(64),
            nn.ReLU(), 
            nn.MaxPool2d(2), 
        )
        self.ConvLayer2 = nn.Sequential(
            nn.Conv2d(64, 128, 3), 
            nn.BatchNorm2d(128), 
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.ConvLayer3 = nn.Sequential(
            nn.Conv2d(128, 256, 3), 
            nn.BatchNorm2d(256),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.ConvLayer4 = nn.Sequential(
            nn.Conv2d(256, 512, 3), 
            nn.BatchNorm2d(512),
            nn.ReLU(),
            nn.MaxPool2d(2)
        )
        self.Linear1 = nn.Sequential(nn.Linear(512 * 11 * 9, 1024), nn.ReLU())
        self.Linear2 = nn.Sequential(nn.Linear(1024, 256), nn.ReLU())
        self.Linear3 = nn.Sequential(nn.Linear(256, 40), nn.Sigmoid())
    def forward(self,x,y):        
        x = self.ConvLayer1(x)
        x = self.ConvLayer2(x)
        x = self.ConvLayer3(x)
        x = self.ConvLayer4(x)
        x = x.view(x.shape[0],-1)

        pred = self.Linear3(self.Linear2(self.Linear1(x))).squeeze(1)

        loss = F.binary_cross_entropy(pred, y, reduction='none')

        return loss, pred


device = 'cuda'
model = Classification_Model()
model.to(device)
optimizer = optim.SGD(model.parameters(),1e-3)

train_set = CELEBA(True,random=True)
train_loader = DataLoader(train_set, batch_size=64, shuffle=True, num_workers=8)
test_set = CELEBA(False,random=True)
test_loader = DataLoader(test_set, batch_size=64, shuffle=False, num_workers=8)

def train_epoch():
    model.train()
    epoch_loss = []
    epoch_pred = []
    epoch_label = []

    for img, label in tqdm(train_loader,ncols=75,leave=False):
        img, label = img.to(device), label.type(torch.FloatTensor).to(device)
        optimizer.zero_grad()
        loss, pred = model(img,label)
        loss.mean().backward()
        optimizer.step()

        epoch_loss.append(loss.to('cpu').data)
        epoch_pred.append(pred.to('cpu').data)
        epoch_label.append(label.to('cpu').data)

    epoch_loss = torch.cat(epoch_loss,0).mean()
    epoch_pred = torch.cat(epoch_pred,0)
    epoch_label = torch.cat(epoch_label,0)
    epoch_auc = roc_auc_score(epoch_label,epoch_pred,average=None)

    return epoch_loss, epoch_auc

def test_epoch():
    model.eval()
    epoch_loss = []
    epoch_pred = []
    epoch_label = []

    for img, label in tqdm(test_loader,ncols=75,leave=False):
        img, label = img.to(device), label.type(torch.FloatTensor).to(device)
        loss, pred = model(img,label)

        epoch_loss.append(loss.to('cpu').data)
        epoch_pred.append(pred.to('cpu').data)
        epoch_label.append(label.to('cpu').data)

    epoch_loss = torch.cat(epoch_loss,0).mean()
    epoch_pred = torch.cat(epoch_pred,0)
    epoch_label = torch.cat(epoch_label,0)
    epoch_auc = roc_auc_score(epoch_label,epoch_pred,average=None)

    return epoch_loss, epoch_auc

if mode=='training':
    for epoch in range(50):
        print('Epoch: ',epoch)
        train_loss, train_auc = train_epoch()
        print('>>>> Train loss: %.5f,   Train Auc: '%(train_loss), train_auc)
        test_loss, test_auc = test_epoch()
        print('>>>> Test loss: %.5f,   Test Auc: '%(test_loss), test_auc)
        print()
        torch.save(model.state_dict(),'saved/classify_random.pth')
else:
    model.load_state_dict(torch.load('saved/classify_random.pth'))
    test_loss, test_auc = test_epoch()
    print('>>>> Test loss: %.5f,   Test Auc: %.5f'%(test_loss,np.mean(test_auc)), test_auc)
    print()
