import torch
import torch.nn as nn
import pdb
import torch.nn.functional as F
from models.transformer_modules import *

class TransformerBlock(nn.Module):
    def __init__(self,in_dim,out_dim,num_heads=4):
        super(TransformerBlock, self).__init__()
        self.num_heads = num_heads
        self.out_dim = out_dim
        self.W = nn.Linear(in_dim,out_dim*3,bias=False)
        self.ffw = nn.Sequential(nn.Linear(out_dim,out_dim),
                                nn.ReLU(),
                                nn.Linear(out_dim,out_dim))
    def forward(self,x,mask=None):
        ### Multihead
        head_outputs = []
        head_inputs = self.W(x)
        head_inputs = torch.split(head_inputs,int(self.out_dim/self.num_heads),2)

        for head_id in range(self.num_heads):
            Q = head_inputs[head_id*3]
            K = head_inputs[head_id*3+1]
            V = head_inputs[head_id*3+2]

            K_transpose = torch.transpose(K,1,2)
            att = torch.bmm(Q,K_transpose)
            att = torch.sigmoid(att)
            if mask is not None:
                att = att * mask.view(mask.shape[0],mask.shape[2],mask.shape[1])
            att = att / (torch.sum(att,2,keepdims=True)+1e-10)

            head_outputs.append(torch.bmm(att,V))

        head_outputs = torch.cat(head_outputs,2)

        head_outputs = head_outputs + x

        return self.ffw(head_outputs) + head_outputs






#### Decoder: CNP to reconstruct original dataset
class DeterministicEncoder(nn.Module):
    def __init__(self, num_layers, x_dim, y_dim, hidden_dim, CNP_mode='deepset'):

        super(DeterministicEncoder, self).__init__()
        self._num_layers = num_layers
        self._CNP_mode = CNP_mode

        if self._CNP_mode == 'deepset':
        	self._module = []
	        for i in range(num_layers-1):
	            if i == 0:
	                in_dim = x_dim + y_dim
	            else:
	                in_dim = hidden_dim
	            self._module.append(nn.Linear(in_dim,hidden_dim))
	            self._module.append(nn.ReLU())
	        self._module.append(nn.Linear(hidden_dim,hidden_dim))
	        self._module = nn.Sequential(*self._module)
        elif self._CNP_mode == 'transformer':
            self._module_dict = nn.ModuleDict()
            for i in range(num_layers):
                if i == 0:
                    in_dim = x_dim + y_dim
                    self._module_dict[str(i)] = nn.Linear(in_dim,hidden_dim)
                else:
                    self._module_dict[str(i)] = TransformerBlock(hidden_dim,hidden_dim)

        else:
            raise Exception('CNP_mode %s: Not implemented'%(self._CNP_mode))

    def forward(self, context_x, context_y, mask=None):

        # Concatenate x and y along the filter axes
        encoder_input = torch.cat([context_x, context_y], axis=-1)

        if self._CNP_mode == 'deepset':
	        # Pass through MLP
	        hidden = self._module(encoder_input)

        elif self._CNP_mode == 'transformer':
            hidden = self._module_dict['0'](encoder_input)
            for i in range(1,self._num_layers):
                hidden = self._module_dict[str(i)](hidden,mask)


        # Aggregator: take the mean over all points
        if mask is not None:
            representation = torch.sum(hidden*mask, axis=1) / (torch.sum(mask,1) + 1e-10)
        else:
            representation =  torch.mean(hidden, axis=1)

        return representation


class DeterministicDecoder(nn.Module):
    def __init__(self, num_layers, x_dim, y_dim, hidden_dim):
        super(DeterministicDecoder, self).__init__()
        self._num_layers = num_layers
        self._x_dim = x_dim 
        self._y_dim = y_dim
        self._module = []
        for i in range(num_layers):
            if i == 0:
                in_dim = x_dim + hidden_dim
            else:
                in_dim = hidden_dim
            self._module.append(nn.Linear(in_dim,hidden_dim))
            self._module.append(nn.ReLU())

        self._module.append(nn.Linear(hidden_dim,y_dim*2))

        self._module = nn.Sequential(*self._module)

    def forward(self, representation, target_x):

        # Concatenate the representation and the target_x
        representation = torch.unsqueeze(representation,1)
        representation = representation.repeat(1, target_x.shape[1], 1)

        input = torch.cat([representation, target_x],-1)


        # Pass through MLP
        hidden = self._module(input)


        # Get the mean an the variance
        mu, log_sigma = torch.split(hidden, self._y_dim, -1)
        # Bound the variance
        sigma = 0.1 + 0.9 * F.softplus(log_sigma)

        # Get the distribution
        dist = torch.distributions.normal.Normal(loc=mu, scale=sigma)

        return dist, mu, sigma


class DeterministicCNP(nn.Module):
    def __init__(self, encoder_num_layers, decoder_num_layers, x_dim, y_dim, hidden_dim, CNP_mode='deepset'):
        super(DeterministicCNP, self).__init__()
        self._encoder = DeterministicEncoder(encoder_num_layers, x_dim, y_dim, hidden_dim, CNP_mode)
        self._decoder = DeterministicDecoder(decoder_num_layers, x_dim, y_dim, hidden_dim)

    def forward(self, context_x, context_y, target_x, target_y=None, mask=None):

        # Pass query through the encoder and the decoder
        representation = self._encoder(context_x, context_y, mask)
        dist, mu, sigma = self._decoder(representation, target_x)

        # If we want to calculate the log_prob for training we will make use of the
        # target_y. At test time the target_y is not available so we return None
        if target_y is not None:
            log_p = dist.log_prob(target_y)
        else:
            log_p = None

        return log_p, mu, sigma






