from __future__ import absolute_import, division, print_function, unicode_literals
import torch
import torch.nn as nn
import torch.optim as optim
from torch.utils.data import DataLoader
import numpy as np
#import matplotlib as mpl
#mpl.use('Agg')
#import matplotlib.pyplot as plt
import collections
from tqdm import tqdm
import pdb
from datetime import datetime
from collections import OrderedDict
import importlib

DATASET = 'function'
MODEL_NAME = 'stochastic_subset'
MODE = 'training'                   # training or inference
INFERENCE_MODE = 'sample'

CHECKPOINT = 'saved/stochastic_subset_function79_run1.pth'

NUM_EPOCHS = 100
BATCH_SIZE = 64
PRETRAIN_CNP = False
PRETRAIN_CHECKPOINT = ''
VALIDATE_WHILE_TRAINING = False


# Model config
config = lambda: None

config.CNP_mode = 'transformer'

config.hidden_dim = 128
config.CNP_encoder_num_layers = 4
config.CNP_decoder_num_layers = 2

# Used for variational subset 
config.subset_encoder_num_layers = 3

# Use for learning to sample
config.sampler_num_conv_layers = 4
config.sampler_num_fc_layers = 4

config.train_with_real_mask = False #only for stochastic_subset


if DATASET == 'function':
    from function import *
    minibatch_per_epoch = 2000

    train_set = FunctionDataSet(minibatch_per_epoch*BATCH_SIZE)
    train_loader = DataLoader(train_set, batch_size=BATCH_SIZE, num_workers=16)

    test_set = FunctionDataSet(200)
    test_loader = DataLoader(test_set, batch_size=1, num_workers=16)
    config.x_dim = 1
    config.y_dim = 1
    config.max_output_points = 30
    config.num_points_jump = 5
    config.max_output_points = 60

elif DATASET == 'kmnist':
    from kmnist import *
    train_set = KMNISTSET(train=True)
    train_loader = DataLoader(train_set, batch_size=BATCH_SIZE, shuffle=True, num_workers=16)

    test_set = KMNISTSET(train=False)
    test_loader = DataLoader(test_set, batch_size=BATCH_SIZE, shuffle=False, num_workers=16)
    config.x_dim = 2
    config.y_dim = 1
    VALIDATE_WHILE_TRAINING = True
    config.max_output_points = 30
    config.num_points_jump = 5
elif DATASET == 'celeba':
    from celeba import *
    train_set = CELEBA(train=True)
    train_loader = DataLoader(train_set, batch_size=BATCH_SIZE, shuffle=True, num_workers=16)

    test_set = CELEBA(train=False)
    test_loader = DataLoader(test_set, batch_size=BATCH_SIZE, shuffle=False, num_workers=16)
    config.x_dim = 2
    config.y_dim = 3
    VALIDATE_WHILE_TRAINING = True
    config.max_output_points = 100
    config.num_points_jump = 20
else:
    raise Exception('There is no dataset '+DATASET)



device = 'cuda'

# Define the model
model_class = getattr(importlib.import_module('models.'+MODEL_NAME), 'DataParallel_Model')
model = model_class(config)
model.to(device)



model_name = '%s_%s'%(MODEL_NAME,DATASET)


# Set up the optimizer and train step
optimizer = optim.Adam(model.parameters(),1e-3)
lr_scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer=optimizer, gamma=0.96)
if MODE == 'training':

    def train_step(x=None, y=None):
        model.train()

        x = x.to(device)
        y = y.to(device)

        optimizer.zero_grad()

        loss, train_dict, _ = model(x,y)
        loss.backward()
        optimizer.step()

        return train_dict



    def test_step(x=None, y=None):
        model.eval()

        x = x.to(device)
        y = y.to(device)
        _, train_dict, test_dict = model(x, y)

        return train_dict, test_dict



    total_dict = OrderedDict()

    for epoch in range(NUM_EPOCHS):
        print('\nEpoch %d:'% (epoch), model_name)
        print('\t Training:')
        for it, (x, y) in tqdm(enumerate(train_loader),total=len(train_loader),ncols=75,leave=False):
            train_dict = train_step(x,y)
            for key in train_dict:
                if key in total_dict:
                    total_dict[key].append(train_dict[key])
                else:
                    total_dict[key] = [train_dict[key]]



        for key in total_dict.keys():
            print('\t%s: %.4f'%(key,np.mean(total_dict[key])))
            total_dict[key] = []


        torch.save(model.state_dict(), 'saved/'+model_name+str(epoch+1)+'_morepoint_run1.pth')
        if VALIDATE_WHILE_TRAINING:
            total_dict = OrderedDict()
            print('\t Validate:')
            for it, (x, y) in tqdm(enumerate(test_loader),total=len(test_loader),ncols=75,leave=False):
                    train_dict,_ = test_step(x,y)
                    for key in train_dict:
                        if key in total_dict:
                            total_dict[key].append(train_dict[key])
                        else:
                            total_dict[key] = [train_dict[key]]

            for key in total_dict.keys():
                print('\t%s: %.4f'%(key,np.mean(total_dict[key])))
                total_dict[key] = []

        lr_scheduler.step()

elif MODE == 'inference':
    model_dict = torch.load(CHECKPOINT)
    model.load_state_dict(model_dict)

    
    def inference_step(x=None, y=None):
        model.eval()

        x = x.to(device)
        y = y.to(device)

        _, train_dict, test_dict = model(x,y)


        return train_dict


    total_dict = OrderedDict()
    print('\nInference:')
    for it, (x, y) in tqdm(enumerate(test_loader),total=len(test_loader),ncols=75,leave=False):
            train_dict = inference_step(x,y)
            for key in train_dict:
                if key in total_dict:
                    total_dict[key].append(train_dict[key])
                else:
                    total_dict[key] = [train_dict[key]]

    for key in total_dict.keys():
        print('\t%s: %.4f'%(key,np.mean(total_dict[key])))
        total_dict[key] = []


